/*

xsample - extended sample objects for Max/MSP and pd (pure data)

Copyright (c) 2001-2011 Thomas Grill (xovo@gmx.net)
For information on usage and redistribution, and for a DISCLAIMER OF ALL
WARRANTIES, see the file, "license.txt," in this distribution.  

$LastChangedRevision: 39 $
$LastChangedDate: 2008-01-04 17:24:49 +0100 (Fr, 04 Jan 2008) $
$LastChangedBy: thomas $
*/

#define FLEXT_ATTRIBUTES 1 

#include <flext.h>

#include <math.h>
